package net.knarcraft.knarlib.particle;

import net.knarcraft.knarlib.util.ParticleHelper;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.World;
import org.bukkit.block.Block;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Collection;
import java.util.Map;
import java.util.UUID;
import java.util.function.Supplier;

/**
 * A runnable tasks that spawns particles at the given blocks
 */
@SuppressWarnings("unused")
public final class ParticleSpawner implements Runnable {

    private final ParticleConfig particleConfig;
    private final Map<Material, ParticleConfig> materialConfigs;

    private Block processingBlock;
    private final Collection<Block> blocks;
    private final Supplier<Collection<Block>> blockSupplier;
    private final UUID storedCalculationId;

    /**
     * Instantiates a new particle spawner
     *
     * @param particleConfig  <p>The configuration for the particle to spawn</p>
     * @param materialConfigs <p>Extra particle configurations for specific materials</p>
     * @param blocks          <p>The blocks to spawn particles on</p>
     */
    public ParticleSpawner(@NotNull ParticleConfig particleConfig,
                           @Nullable Map<Material, ParticleConfig> materialConfigs, @NotNull Collection<Block> blocks) {
        this.particleConfig = particleConfig;
        this.materialConfigs = materialConfigs;
        this.blocks = blocks;
        this.blockSupplier = null;

        this.storedCalculationId = UUID.randomUUID();
    }

    /**
     * Instantiates a new particle spawner
     *
     * @param particleConfig  <p>The configuration for the particle to spawn</p>
     * @param materialConfigs <p>Extra particle configurations for specific materials</p>
     * @param blockSupplier   <p>The supplier supplying the blocks to spawn particles on</p>
     */
    public ParticleSpawner(@NotNull ParticleConfig particleConfig,
                           @Nullable Map<Material, ParticleConfig> materialConfigs,
                           @NotNull Supplier<Collection<Block>> blockSupplier) {
        this.particleConfig = particleConfig;
        this.materialConfigs = materialConfigs;
        this.blocks = null;
        this.blockSupplier = blockSupplier;

        this.storedCalculationId = UUID.randomUUID();
    }

    /**
     * Gets the id used for stored calculations
     *
     * <p>You should run ParticleHelper.clearStoredCalculations(id) with this id when discarding this particle spawner
     * to prevent a memory leak.</p>
     *
     * @return <p>The id used for stored calculations</p>
     */
    public @NotNull UUID getStoredCalculationId() {
        return this.storedCalculationId;
    }

    @Override
    public void run() {
        // Use the static collection of blocks, or a dynamically changing supplier
        Collection<Block> blocksToSpawnOn;
        if (blocks != null) {
            blocksToSpawnOn = blocks;
        } else if (blockSupplier != null) {
            blocksToSpawnOn = blockSupplier.get();
        } else {
            throw new RuntimeException("There is a bug in the plugin code. Please contact the developer!");
        }

        for (Block block : blocksToSpawnOn) {
            // Ignore blocks in unloaded chunks
            if (!block.getChunk().isLoaded()) {
                continue;
            }

            Location location = block.getLocation().clone();
            World world = location.getWorld();
            if (world == null) {
                continue;
            }

            spawnParticleForBlock(block, world, location);
        }
    }

    /**
     * Spawns the defined particle for the given block
     *
     * @param block    <p>The block to spawn a particle effect for</p>
     * @param world    <p>The world the block belongs to</p>
     * @param location <p>A clone of the block's location</p>
     */
    private void spawnParticleForBlock(@NotNull Block block, @NotNull World world, @NotNull Location location) {
        // Store the currently processed block for height calculation
        processingBlock = block;

        double blockHeight = ParticleHelper.getBlockHeight(block);
        ParticleConfig activeConfig = getParticleConfig();

        switch (getParticleConfig().getParticleMode()) {
            case SINGLE -> ParticleHelper.spawnParticle(world, location.clone().add(0.5,
                    activeConfig.getHeightOffset(), 0.5), activeConfig, blockHeight);
            case SQUARE -> ParticleHelper.drawSquare(world, location, activeConfig, blockHeight);
            case CIRCLE -> ParticleHelper.drawCircle(world, location, activeConfig, blockHeight, storedCalculationId);
            case PYRAMID -> ParticleHelper.drawPyramid(world, location, activeConfig, blockHeight, storedCalculationId);
            case SPHERE -> ParticleHelper.drawSphere(world, location, activeConfig, blockHeight, storedCalculationId);
            case CUBE -> ParticleHelper.drawCube(world, location, activeConfig, blockHeight);
        }
    }

    /**
     * Gets the particle config to use for the current block's material
     *
     * @return <p>The particle config to use</p>
     */
    private @NotNull ParticleConfig getParticleConfig() {
        if (this.materialConfigs == null) {
            return this.particleConfig;
        }
        ParticleConfig materialConfig = this.materialConfigs.get(processingBlock.getType());
        if (materialConfig != null) {
            return materialConfig;
        }
        return this.particleConfig;
    }

}
